import {IdentifierType, Model, ModelShape, SerializedModel} from "@sharkitek/core";
import {eq, getTableColumns, Table, TableConfig} from "drizzle-orm";
import {PgDatabase} from "drizzle-orm/pg-core";
import {PgQueryResultHKT} from "drizzle-orm/pg-core/session";
import type {ExtractTablesWithRelations, TablesRelationalConfig} from "drizzle-orm/relations";
import {ModelQuery} from "./Query";

/**
 * Serialized values converters for database.
 */
export const serializedToDatabaseTypes: Record<string, (serializedValue: any) => any> = {
	"date": (serializedValue: string) => new Date(serializedValue),
};

/**
 * Sharkitek model extension for Drizzle.
 */
export interface DrizzleExtension<
	ModelType extends Model<Shape, IdentifierType<Shape, Identifier>>,
	TableType extends Table<TC>, TC extends TableConfig,
	TQueryResult extends PgQueryResultHKT, TFullSchema extends Record<string, unknown>,
	Scopes extends object,
	Shape extends ModelShape, Identifier extends keyof Shape = any, TSchema extends TablesRelationalConfig = ExtractTablesWithRelations<TFullSchema>
>
{
	/**
	 * Drizzle model manager.
	 */
	drizzle(): DrizzleModel<ModelType, TableType, TC, TQueryResult, TFullSchema, Scopes, Shape, Identifier, TSchema>;
}

/**
 * Drizzle model manager.
 */
export class DrizzleModel<
	ModelType extends Model<Shape, IdentifierType<Shape, Identifier>>,
	TableType extends Table<TC>, TC extends TableConfig,
	TQueryResult extends PgQueryResultHKT, TFullSchema extends Record<string, unknown>,
	Scopes extends object,
	Shape extends ModelShape, Identifier extends keyof Shape = any, TSchema extends TablesRelationalConfig = ExtractTablesWithRelations<TFullSchema>
>
{
	constructor(protected model: ModelType,
	            protected database: PgDatabase<TQueryResult, TFullSchema, TSchema>,
	            protected table: TableType,
	            protected scopes: Scopes & ThisType<ModelQuery<ModelType, TableType, TC, TQueryResult, TFullSchema, Shape, Identifier, TSchema>>)
	{}

	/**
	 * Get Drizzle database.
	 */
	getDatabase(): PgDatabase<TQueryResult, TFullSchema, TSchema>
	{
		return this.database;
	}

	/**
	 * Get model Drizzle table.
	 */
	getTable(): TableType
	{
		return this.table;
	}

	/**
	 * Get scopes.
	 */
	getScopes(): Scopes & ThisType<ModelQuery<ModelType, TableType, TC, TQueryResult, TFullSchema, Shape, Identifier, TSchema>>
	{
		return this.scopes;
	}

	/**
	 * Save the model in database.
	 */
	async save(): Promise<boolean>
	{
		// Get serialized model update.
		const serializedModel = this.model.patch();

		// Get table columns of the model.
		const tableColumns = getTableColumns(this.table);

		// Create model data for the database.
		const databaseModelData = Object.fromEntries(
			Object.entries(serializedModel).map(
				// Only keep table column values and convert them into database types, if it is required.
				([key, value]) => [key, tableColumns?.[key] ? (serializedToDatabaseTypes?.[tableColumns?.[key].dataType]?.(value) ?? value) : undefined]
			)
		) as Partial<SerializedModel<Shape>>;

		// Initialize query result variable.
		let result: any;

		if (this.model.isNew())
		{ // Insert the new model in database and get the inserted row data.
			result = (await (this.database.insert(this.table)).values(databaseModelData as any).returning() as any)?.[0];
		}
		else
		{ // Update model data in database, and get the updated row data.
			result = (await (this.database.update(this.table)
				.set(databaseModelData)
				.where(eq((this.table as any)?.[this.model.getIdentifierName()], this.model.getIdentifier())))
				.returning() as any)?.[0];
		}

		// Update model from inserted or updated row data.
		this.model.deserialize(result);

		return true;
	}

	async refresh(): Promise<void>
	{
		// Update model from up-to-date row data.
		this.model.deserialize(
			(( // Retrieve model data from database.
				await this.database.select().from(this.table)
					.where(eq((this.table as any)?.[this.model.getIdentifierName()], this.model.getIdentifier()))
			) as any)[0]
		);
	}
}

/**
 * Drizzle extension initializer.
 * @param database Drizzle database.
 * @param table Drizzle database table.
 * @param scopes
 */
export function drizzle<
	ModelType extends Model<Shape, IdentifierType<Shape, Identifier>>,
	TableType extends Table<TC>, TC extends TableConfig,
	TQueryResult extends PgQueryResultHKT, TFullSchema extends Record<string, unknown>,
	Scopes extends object,
	Shape extends ModelShape, Identifier extends keyof Shape = any, TSchema extends TablesRelationalConfig = ExtractTablesWithRelations<TFullSchema>,
>(
	database: PgDatabase<TQueryResult, TFullSchema, TSchema>,
	table: TableType,
	scopes: Scopes & ThisType<ModelQuery<ModelType, TableType, TC, TQueryResult, TFullSchema, Shape, Identifier, TSchema>>
): DrizzleExtension<ModelType, TableType, TC, TQueryResult, TFullSchema, Scopes, Shape, Identifier, TSchema> & ThisType<ModelType>
{
	// Return initialized drizzle extension.
	return {
		drizzle(): DrizzleModel<ModelType, TableType, TC, TQueryResult, TFullSchema, Scopes, Shape, Identifier, TSchema>
		{
			// Initialize drizzle model manager instance.
			return new DrizzleModel<ModelType, TableType, TC, TQueryResult, TFullSchema, Scopes, Shape, Identifier, TSchema>(this, database, table, scopes);
		}
	};
}
